(*@ function rec assoc (v1: int) (suc: (int * int list) list) :
      int list
    = match suc with
      | [] -> []
      | (v, l) :: rs -> if v1 = v then l else assoc v1 rs *)

(*@ function rec append_last (x: 'a) (l: 'a list) : 'a list =
      match l with
      | [] -> x :: []
      | y :: r -> y :: append_last x r *)

type 'a queue
(*@ mutable model contents : 'a list *)

val create : unit -> 'a queue
(*@ t = create ()
      ensures t.contents = [] *)

val is_empty : 'a queue -> bool
(*@ b = is_empty t
      ensures b = if t.contents = [] then true
                  else false *)

val push : 'a -> 'a queue -> unit
(*@ push x t
      modifies t.contents
      ensures t.contents = append_last x (old t.contents) *)

exception Empty

val pop : 'a queue -> 'a
(*@ x = pop t
      modifies t.contents
      ensures  t.contents = if old t.contents = []
                            then []
                            else List.tl (old t.contents)
      ensures  if old t.contents = [] then false
               else x = List.hd (old t.contents)
      raises Empty -> old t.contents = [] = t.contents *)

type t
(*@ mutable model dom : int list
    mutable model succ : (int * int list) list *)

val empty : unit -> t
(*@ t = empty ()
      ensures t.dom = []
      ensures t.succ = [] *)

val add_vertex : int -> t -> unit
(*@ add_vertex v t
      modifies t.dom
      ensures t.dom = v :: (old t.dom) *)

val add_edge : int -> int -> t -> unit
(*@ add_edge v1 v2 t
      modifies t.dom, t.succ
      ensures t.dom = v1 :: v2 :: (old t.dom)
      ensures t.succ = (v1, v2 :: (assoc v1 (old t.succ))) :: (old t.succ) *)

(* @ predicate edge (v1: int) (v2: int) (t: t) =
      List.mem v2 (assoc v1 t.succ) *)

(* @ predicate rec path (v1: int) (l: int list) (v2: int) (t: t) =
      match l with
      | [] -> v1 = v2
      | x :: rs -> List.mem v1 t.dom && edge v1 x t && path x rs v2 t *)

(* @ predicate is_path (v1 : G.V.t) (l : G.V.t seq) (v2 : G.V.t) (g : G.gt) =
      let len = Seq.length l in
      if len = 0 then v1 = v2 else
        edge v1 l[0] g && l[len - 1] = v2 && Fset.mem v1 g.G.dom &&
      forall i : int. 0 <= i < len - 1 -> edge l[i] l[i+1] g *)

val check_path : int queue -> int -> int -> t -> bool
(*@ b = check_path q v1 v2 t
      requires List.mem v1 t.dom
      raises Empty -> q.contents = []
      (* ensures  b <-> exists l. path v1 l v2 t *) *)

(*
 * Local Variables:
 * compile-command: "ortac qcheck-stm path.mli "empty ()" "t" > ../bin/main.ml"
 * End:
 *)
